DEF UV_STREAM_RECV_BUF_SIZE = 256000  # 250kb

DEF FLOW_CONTROL_HIGH_WATER = 65536
DEF FLOW_CONTROL_LOW_WATER = 16384

DEF MAX_THREADPOOL_WORKERS = 5

DEF DEFAULT_FREELIST_SIZE = 250


DEF MAX_SIG = 64


DEF __PROCESS_DEBUG_SLEEP_AFTER_FORK = 1
